require_relative 'blog_api'

RSpec.describe ApiResource::Resource do
  before do
    @blog          = { data: { id: 1257, title: 'Tarte a la creme' } }
    @tags_data     = [{ id: 33, name: 'food' }, { id: 39, name: 'cuisine' }]
    @category_data = { id: 33, name: 'Hobbies' }
  end

  context 'has_many' do
    def check_has_many(relation_name, relation_values, aggregate_class=BlogApi::Blog, component_class=BlogApi::Tag)
      resource_values = @blog.merge(data: { relation_name => relation_values })
      resource_id     = 3
      req(:get, "/#{aggregate_class.resource_path}/#{resource_id}", resource_values)
      blog = aggregate_class.find(resource_id)
      check_returned_object(aggregate_class, resource_values, blog, relation_name)
      expect(blog).to respond_to(relation_name)
      check_returned_array(component_class, relation_values, blog.send(relation_name))
      yield blog if block_given?
    end

    it 'guess type from field name for `data` rooted array of data' do
      check_has_many(:tags, data: @tags_data)
    end

    it 'guess type from field name for plain array of data' do
      check_has_many(:tags, @tags_data)
    end

    it 'override type with meta.type and retrieve other meta.* info' do
      check_has_many(:my_sweet_tags,
                     data: [{ id: 33, name: 'food' }, { id: 39, name: 'cuisine' }],
                     meta: { type: :tags, page: 3, per_page: 50, total: 102 }) do |blog|
        expect(blog.my_sweet_tags.page).to eq(3)
        expect(blog.my_sweet_tags.per_page).to eq(50)
        expect(blog.my_sweet_tags.total).to eq(102)
      end
    end

    context '::by_parent called on aggregate class' do
      it 'guess type from field name for `data` rooted array of data' do
        check_has_many(:tags, { data: @tags_data }, BlogApi::Blog.by_author('Flinn'))
      end

      it 'guess type from field name for plain array of data' do
        check_has_many(:tags, @tags_data, BlogApi::Blog.by_author('Flinn'))
      end

      it 'override type with meta.type and retrieve other meta.* info' do
        check_has_many(:my_sweet_tags,
                       { data: [{ id: 33, name: 'food' }, { id: 39, name: 'cuisine' }],
                         meta: { type: :tags, page: 3, per_page: 50, total: 102 } },
                       BlogApi::Blog.by_author('Flinn')) do |blog|
          expect(blog.my_sweet_tags.page).to eq(3)
          expect(blog.my_sweet_tags.per_page).to eq(50)
          expect(blog.my_sweet_tags.total).to eq(102)
        end
      end
    end
  end

  context 'has_one' do
    let(:raw_data) { { 'foo' => 'bar', 'baz' => 3 } }

    def check_has_one(relation_name, relation_values, relation_type=BlogApi::Category)
      @blog[:data].merge!(relation_name => relation_values)
      resource_id = 3
      stub        = req(:get, "/blogs/#{resource_id}", @blog)
      blog        = BlogApi::Blog.find(resource_id)
      check_returned_object(BlogApi::Blog, @blog, blog, relation_name)
      expect(blog).to respond_to(relation_name)
      check_returned_object(relation_type, relation_values, blog.send(relation_name))
      expect(stub).to have_been_requested
    end

    it 'gets data rooted inlined json' do
      check_has_one(:category, data: @category_data)
    end

    it 'gets un-rooted inlined json' do
      check_has_one(:category, @category_data)
    end

    it 'override type from meta' do
      check_has_one(:main_category, data: { id: 33, name: 'Hobbies' }, meta: { type: :category })
    end

    it 'creates method for raw types' do
      check_has_one(:raw_attr, raw_data, Hash)
    end

    it 'replaces method with raw types' do
      check_has_one(:do_stuff_method, raw_data, Hash)
    end
  end
end