require_relative 'blog_api'

RSpec.describe ApiResource::Resource do
  before do
    @expected_resources = { data: [{ id: 1257, title: 'Hello' },
                                   { id: 33, title: 'World' },
                                   { id: 38, title: 'Bonjour!' }] }
  end

  context '::all_pages' do
    it 'retrieves all pages' do
      expected_values = (1..223).map { |i| { id: i, title: "Hello #{i}" } }.each_slice(100).to_a
      expected_values.each_with_index do |objs, i|
        req(:get, "/blogs?page=#{ i+1 }&per_page=100", { data: objs, meta: { total_count: 223 } })
      end

      i = 1
      BlogApi::Blog.all_pages(per_page: 100) do |elts, page, per_page, total_count|
        expect(total_count).to eq(223)
        expect(page).to eq(i)
        expect(per_page).to eq(100)
        check_returned_array(BlogApi::Blog, expected_values[i-1], elts)
        i += 1
        false
      end
      expect(i).to eq(4)
    end
  end

  context '::find' do
    let(:blog_attrs) { { id: 1257, title: 'Hello' } }
    let(:extra_params) { { foo: 3, bar: 'baz' } }

    def check_found_resource(returned_data, extra_params={})
      resource_id = blog_attrs[:id]
      stub        = req(:get, "/blogs/#{resource_id}?#{extra_params.to_param}", returned_data)
      result      = BlogApi::Blog.find(resource_id, extra_params)
      check_returned_object(BlogApi::Blog, returned_data, result)
      expect(stub).to have_been_requested
    end

    it 'creates resource from data rooted json' do
      check_found_resource(data: blog_attrs)
      check_found_resource({ data: blog_attrs }, extra_params)
    end

    it 'creates resource from un-rooted json' do
      check_found_resource(blog_attrs)
      check_found_resource(blog_attrs, extra_params)
    end
  end

  context '::all' do

    it 'fetches all resources' do
      req(:get, '/blogs', @expected_resources)
      result = BlogApi::Blog.all
      check_returned_array(BlogApi::Blog, @expected_resources, result)
    end
  end

  context '::where' do
    let(:query_params) { { title: 'hello world', tags: ['hello', 'first', 'greeting'] } }
    it 'fetches resourced filtered by query string with GET' do
      req(:get, "/blogs?#{query_params.to_param}", @expected_resources)
      result = BlogApi::Blog.where(query_params)
      check_returned_array(BlogApi::Blog, @expected_resources, result)
    end

    it 'not add query string for empty array parameter values on GET' do
      req(:get, '/blogs', @expected_resources)
      result = BlogApi::Blog.where({ tags: [] })
      check_returned_array(BlogApi::Blog, @expected_resources, result)
    end
  end

  context '#by_parent_resource' do

    it 'fetches resource nested with parent resource' do
      resource_id = 3
      req(:get, "/tags/#{resource_id}/blogs", @expected_resources)
      result = BlogApi::Blog.by_tag(resource_id).all
      check_returned_array(BlogApi::Blog, @expected_resources, result)
    end
  end

  context 'empty base_url' do
    it 'throws on requests' do
      class Foo < ApiResource::Resource;
      end
      expect(-> { Foo.all }).to raise_exception(StandardError)
    end
  end

  context '#save!' do
    let(:blog_attrs) { { id: 525, title: 'hello' } }
    let(:blog_id) { blog_attrs[:id] }
    let(:expected_value) { { data: blog_attrs } }

    it 'POSTs data when id is nil' do
      blog     = BlogApi::Blog.new(title: 'hello')
      stub     = req(:post, '/blogs', expected_value, 201, blog_attrs.except(:id))
      returned = blog.save!
      check_returned_object(BlogApi::Blog, expected_value, blog)
      expect(returned).to eq(blog)
      expect(blog.id).to eq(blog_id)
      expect(stub).to have_been_requested
    end

    it 'PUTs data when id is not nil' do
      blog     = BlogApi::Blog.new(id: blog_id, title: 'hello')
      stub     = req(:put, "/blogs/#{blog_id}", nil, 204, blog_attrs.except(:id))
      returned = blog.save!
      expect(returned).to eq(blog)
      expect(blog.id).to eq(blog_id)
      expect(stub).to have_been_requested
    end

    context '#by_resource' do
      it 'POSTs data in parent resource context' do
        blog_id        = 525
        blog           = BlogApi::Blog.new(title: 'hello')
        expected_value = { data: { id: blog_id, title: 'hello' } }
        user_id        = 'ux345'

        stub     = req(:post, "/users/#{user_id}/blogs", expected_value, 201, blog_attrs.except(:id))
        returned = blog.by_user(user_id).save!
        check_returned_object(BlogApi::Blog, expected_value, blog)
        expect(returned).to eq(blog)
        expect(blog.id).to eq(blog_id)
        expect(stub).to have_been_requested
      end

      it 'PUTs data in parent resource context when id is not nil' do
        blog    = BlogApi::Blog.new(id: blog_id, title: 'hello')
        user_id = 'ux345'

        stub     = req(:put, "/users/#{user_id}/blogs/#{blog_id}", nil, 204, blog_attrs.except(:id))
        returned = blog.by_user(user_id).save!
        expect(returned).to eq(blog)
        expect(blog.id).to eq(blog_id)
        expect(stub).to have_been_requested
      end
    end
  end

  context '#destroy!' do
    it 'DELETEs data when id is not nil' do
      blog = BlogApi::Blog.new(id: 323, title: 'hello')
      stub = req(:delete, "/blogs/#{blog.id}", '', 204)

      returned = blog.destroy!
      expect(returned).to eq(blog)
      expect(stub).to have_been_requested
    end
  end

  context '#submit!' do

    class SignUp < BlogApi::Resource
      self.resource_path = 'sign_up'
      attr_accessor :email, :password, :password_confirmation, :name
      validates :name, presence: true
    end

    class User < BlogApi::Resource
      attr_accessor :id, :email, :name
    end

    let(:sign_up_data) { { email: 'user@example.com', password: 'pass', password_confirmation: 'pass', name: 'Joe' } }
    let(:sign_up) { SignUp.new(sign_up_data) }

    it 'uses the resource_path and uses the type in meta' do
      expected_value = { data: { id: 3, email: 'user@example.com', name: 'Jane' }, meta: { type: 'user' } }
      stub           = req(:post, '/sign_up', expected_value, 200, sign_up_data)
      returned       = sign_up.submit!
      check_returned_object(User, expected_value, returned)
      expect(stub).to have_been_requested
    end

    it 'sets the returned errors' do
      expected_value = { errors: { password_confrimation: ["can't be blank"], email: ["can't be blank", 'is invalid'] } }
      stub           = req(:post, '/sign_up', expected_value, 422, sign_up.attributes)
      expect(proc { sign_up.submit! }).to raise_error(ApiResource::ResourceError) do |error|
        expect(error.model).to eq(sign_up)
      end
      expect(sign_up.errors.messages).to match(expected_value[:errors])
      expect(stub).to have_been_requested
    end

    it 'no request if not valid' do
      sign_up.name = nil
      stub         = req(:post, '/sign_up', nil, 422)
      expect(proc { sign_up.submit! }).to raise_error(ApiResource::ResourceError) do |error|
        expect(error.model).to eq(sign_up)
      end
      expect(sign_up.errors[:name]).to be_truthy
      expect(stub).to_not have_been_requested
    end
  end

  context '::filter_submit_params' do
    class Purchase < BlogApi::Resource
      filter_submitted_attributes do |attrs|
        address_attrs = [:name, :street, :zip_code]
        attrs.except(*address_attrs).merge(address: attrs.slice(*address_attrs))
      end
      attr_accessor :amount, :name, :street, :zip_code
    end
    it 'submit! filters params before being submitted' do
      comment = Purchase.new(amount: 30, name: 'Jane', street: 'Old Street', zip_code: 'EC1Y 1BE')
      stub    = req(:post, '/purchases', nil, 200, { amount: '30', address: { name: 'Jane', street: 'Old Street', zip_code: 'EC1Y 1BE' } })
      comment.submit!
      expect(stub).to have_been_requested
    end
  end
end